@@ -6,7 +6,7 @@
  */
 
 /*
-** Copyright © 1993-1998,2001,2006,2021 Marko Mäkelä
+** Copyright © 1993-1998,2001,2006,2021,2022 Marko Mäkelä
 ** 1581 disk image management by Pasi Ojala
 **
 **     This program is free software; you can redistribute it and/or modify
@@ -330,7 +330,7 @@ isFreeBlock (const struct Image* image, byte_t track, byte_t sector)
     if (!(BAM = getBlock ((struct Image*) image, image->dirtrack, 1)))
       return false;
 
-    if(track > 40) {
+    if (track > 40) {
       if (!(BAM = getBlock ((struct Image*) image, BAM[0], BAM[1])))
         return false;
 
@@ -357,6 +357,7 @@ findNextFree (const struct Image* image,
 {
   const struct DiskGeometry* geom;
   byte_t t = *track, s = *sector;
+  size_t visited[64 / (sizeof(size_t) * CHAR_BIT)];
   unsigned i;
 
   if (!image || !image->buf || !(geom = getGeometry (image->type)))
@@ -368,61 +369,111 @@ findNextFree (const struct Image* image,
   if (t >= image->dirtrack) {
     /* search from the current track upwards */
 
-    for (; t <= image->partTops[image->dirtrack - 1]; t++)
+    for (; t <= image->partTops[image->dirtrack - 1]; t++) {
+      memset(visited, 0, sizeof visited);
       for (i = geom->sectors[t]; i; i--) {
         if (isFreeBlock (image, t, s)) {
           *track = t;
           *sector = s;
           return true;
         }
+        visited[s / ((sizeof *visited) * CHAR_BIT)] |=
+          ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT));
         s += geom->interleave[t];
         s %= geom->sectors[t];
+
+        while (visited[s / ((sizeof *visited) * CHAR_BIT)] &
+               ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT)))
+          if (++s == geom->sectors[t]) {
+            s = 0;
+            if (i == 1)
+              break;
+            i--;
+          }
       }
+    }
 
     /* search from lower tracks (from the directory track downwards) */
 
     for (t = image->dirtrack - 1;
-         t >= image->partBots[image->dirtrack - 1]; t--)
+         t >= image->partBots[image->dirtrack - 1]; t--) {
+      memset(visited, 0, sizeof visited);
       for (i = geom->sectors[t]; i; i--) {
         if (isFreeBlock (image, t, s)) {
           *track = t;
           *sector = s;
           return true;
         }
+        visited[s / ((sizeof *visited) * CHAR_BIT)] |=
+          ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT));
         s += geom->interleave[t];
         s %= geom->sectors[t];
+        while (visited[s / ((sizeof *visited) * CHAR_BIT)] &
+               ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT)))
+          if (++s == geom->sectors[t]) {
+            s = 0;
+            if (i == 1)
+              break;
+            i--;
+          }
       }
+    }
   }
   else {
     /* search from the current track downwards */
 
-    for (; t >= image->partBots[image->dirtrack - 1]; t--)
+    for (; t >= image->partBots[image->dirtrack - 1]; t--) {
+      memset(visited, 0, sizeof visited);
       for (i = geom->sectors[t]; i; i--) {
         if (isFreeBlock (image, t, s)) {
           *track = t;
           *sector = s;
           return true;
         }
+        visited[s / ((sizeof *visited) * CHAR_BIT)] |=
+          ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT));
         s += geom->interleave[t];
         s %= geom->sectors[t];
+        while (visited[s / ((sizeof *visited) * CHAR_BIT)] &
+               ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT)))
+          if (++s == geom->sectors[t]) {
+            s = 0;
+            if (i == 1)
+              break;
+            i--;
+          }
       }
+    }
 
     /* search from upper tracks (from the directory track upwards) */
 
     for (t = image->dirtrack + 1;
-         t <= image->partTops[image->dirtrack - 1]; t++)
+         t <= image->partTops[image->dirtrack - 1]; t++) {
+      memset(visited, 0, sizeof visited);
       for (i = geom->sectors[t]; i; i--) {
         if (isFreeBlock (image, t, s)) {
           *track = t;
           *sector = s;
           return true;
         }
+        visited[s / ((sizeof *visited) * CHAR_BIT)] |=
+          ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT));
         s += geom->interleave[t];
         s %= geom->sectors[t];
+        while (visited[s / ((sizeof *visited) * CHAR_BIT)] &
+               ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT)))
+          if (++s == geom->sectors[t]) {
+            s = 0;
+            if (i == 1)
+              break;
+            i--;
+          }
       }
+    }
 
     /* last resort: search from the directory track */
     t = image->dirtrack;
+    memset(visited, 0, sizeof visited);
 
     for (i = geom->sectors[t]; i; i--) {
       if (isFreeBlock (image, t, s)) {
@@ -430,8 +481,18 @@ findNextFree (const struct Image* image,
         *sector = s;
         return true;
       }
+      visited[s / ((sizeof *visited) * CHAR_BIT)] |=
+        ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT));
       s += geom->interleave[t];
       s %= geom->sectors[t];
+      while (visited[s / ((sizeof *visited) * CHAR_BIT)] &
+             ((size_t) 1) << (s % ((sizeof *visited) * CHAR_BIT)))
+        if (++s == geom->sectors[t]) {
+          s = 0;
+          if (i == 1)
+            break;
+          i--;
+        }
     }
   }
 
@@ -704,7 +765,8 @@ FormatImage (struct Image* image)
     BAM[0xA2] = id1;
     BAM[0xA3] = id2;
     /* free all blocks */
-    memset (&BAM[4], 0xFF, (size_t) geom->tracks << 2);
+    memset (&BAM[4], 0xFF, 35 << 2);
+    memset (&BAM[683 << 8], 0xFF, 35 * 3);
 
     for (track = 1; track <= 35; track++) {
       /* set amount of free blocks on each track */
@@ -719,7 +781,8 @@ FormatImage (struct Image* image)
       BAM[0xDC + track + 1] = sector = geom->sectors[track + 1];
       /* allocate non-existent blocks */
       for (; sector < 24; sector++)
-        BAM[(track * 3) + (sector >> 3)] &= (byte_t) ~(1 << (sector & 7));
+        BAM[(683 << 8) + (track * 3) + (sector >> 3)] &=
+          (byte_t) ~(1 << (sector & 7));
     }
 
     /* Allocate the BAM and directory entries. */
@@ -890,7 +953,6 @@ backupBAM (const struct Image* image, byte_t** BAM)
     const byte_t* bamblock;
 
   case ImUnknown:
-    return false;
   case Im1541:
     if (!(bamblock = getBlock ((struct Image*) image, image->dirtrack, 0)))
       return false;
@@ -910,16 +972,9 @@ backupBAM (const struct Image* image, byte_t** BAM)
       return false;
 
     memcpy (*BAM, &bamblock[4], 35 << 2);
-    memcpy (BAM[35 << 2], &bamblock[0xDD], 35);
+    memcpy (*BAM + (35 << 2), &bamblock[0xDD], 35);
 
-    if (!(bamblock = getBlock ((struct Image*) image,
-                               35 + image->dirtrack, 0))) {
-      free (*BAM);
-      *BAM = 0;
-      return false;
-    }
-
-    memcpy (BAM[35 * 5], bamblock, 35 * 3);
+    memcpy (*BAM + 35 * 5, bamblock + (683 << 8), 35 * 3);
     return true;
 
   case Im1581:
@@ -962,7 +1017,7 @@ restoreBAM (struct Image* image, byte_t** BAM)
 
   switch (image->type) {
   case ImUnknown:
-    return false;
+    break;
   case Im1541:
     {
       byte_t* bamblock;
@@ -971,11 +1026,12 @@ restoreBAM (struct Image* image, byte_t** BAM)
         return false;
 
       memcpy (&bamblock[4], *BAM, (size_t) geom->tracks << 2);
-      free (*BAM);
-      *BAM = 0;
-
-      return true;
     }
+  done:
+    free (*BAM);
+    *BAM = 0;
+
+    return true;
   case Im1571:
     {
       byte_t* bamblock;
@@ -984,16 +1040,10 @@ restoreBAM (struct Image* image, byte_t** BAM)
         return false;
 
       memcpy (&bamblock[4], *BAM, 35 << 2);
-      memcpy (&bamblock[0xDD], BAM[35 << 2], 35);
+      memcpy (&bamblock[0xDD], *BAM + (35 << 2), 35);
 
-      if (!(bamblock = getBlock (image, 35 + image->dirtrack, 0)))
-        return false;
-
-      memcpy (bamblock, BAM[35 * 5], 35 * 3);
-      free (*BAM);
-      *BAM = 0;
-
-      return true;
+      memcpy (&bamblock[683 << 8], *BAM + 35 * 5, 35 * 3);
+      goto done;
     }
   case Im1581:
     {
@@ -1008,10 +1058,7 @@ restoreBAM (struct Image* image, byte_t** BAM)
       memcpy (bamblocks[1], *BAM + 256, 256);
 
       free (bamblocks);
-      free (BAM);
-      *BAM = 0;
-
-      return true;
+      goto done;
     }
   }
 
@@ -1578,7 +1625,8 @@ setupSideSectors (struct Image* image,
   {
     byte_t** sidesect = 0;
     byte_t** datafile = 0;
-    byte_t ss, ssentry, i, track, sector;
+    size_t ssentry;
+    byte_t ss, i, track, sector;
 
     if (blocks != mapInode (&datafile, image,
                             dirent->firstTrack, dirent->firstSector,
@@ -1608,12 +1656,11 @@ setupSideSectors (struct Image* image,
     track = dirent->firstTrack;
     sector = dirent->firstSector;
 
-    for (ss = ssentry = 0; track; ssentry++) {
-      ss = ssentry / 120;
-
-      if (ss >= sscount)
+    for (ssentry = 0; track; ssentry++) {
+      if (ssentry / 120 >= sscount)
         return WrFail;
 
+      ss = (byte_t) (ssentry / 120);
       sidesect[ss][16 + (ssentry % 120) * 2] = track;
       sidesect[ss][17 + (ssentry % 120) * 2] = sector;
 
@@ -2736,9 +2783,9 @@ ReadImage (FILE* file,
 
             for (j = 0; j < sizeof name.name; j++)
               if (name.name[j] >= 'A' && name.name[j] <= 'Z')
-                name.name[j] = name.name[j] - 'A' + 0xC1;
+                name.name[j] -= (unsigned char) ('A' + 0xC1);
               else if (name.name[j] >= 'a' && name.name[j] <= 'z')
-                name.name[j] = name.name[j] - 'a' + 0x41;
+                name.name[j] -= (unsigned char) ('a' + 0x41);
 
             name.type = PRG;
           }
